/* 
 * FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 * Copyright (C) 2005/2006, Anthony Minessale II <anthmct@yahoo.com>
 *
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 *
 * The Initial Developer of the Original Code is
 * Anthony Minessale II <anthmct@yahoo.com>
 * Portions created by the Initial Developer are Copyright (C)
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Anthony Minessale II <anthmct@yahoo.com>
 *
 *
 * mod_sipx.c -- SIPX SIP Endpoint
 *
 */



#define MY_EVENT_REGISTER "sipx::register"
#define MY_EVENT_EXPIRE "sipx::expire"

#define HAVE_APR
#include <switch.h>
#include <net/SipConfigServerAgent.h>
#include <net/SipUserAgent.h>

#define DBFILE "sipx"

static const char modname[] = "mod_sipx";
#define STRLEN 15

static switch_memory_pool_t *module_pool = NULL;

#define set_param(ptr,val) if (ptr) {free(ptr) ; ptr = NULL;} if (val) {ptr = strdup(val);}
#define set_anchor(t,m) if (t->Anchor) {delete t->Anchor;} t->Anchor = new SipMessage(m);

#ifndef _64BIT_MACHINE_
typedef int sipx_int_t;
#endif

typedef enum {
	TFLAG_IO = (1 << 0),
	TFLAG_INBOUND = (1 << 1),
	TFLAG_OUTBOUND = (1 << 2),
	TFLAG_READING = (1 << 3),
	TFLAG_WRITING = (1 << 4),
	TFLAG_HUP = (1 << 5),
	TFLAG_RTP = (1 << 6),
	TFLAG_BYE = (1 << 7),
	TFLAG_ANS = (1 << 8),
	TFLAG_EARLY_MEDIA = (1 << 9),
	TFLAG_SECURE = (1 << 10),
	TFLAG_VAD_IN = ( 1 << 11),
	TFLAG_VAD_OUT = ( 1 << 12),
	TFLAG_VAD = ( 1 << 13),
	TFLAG_TIMER = (1 << 14)
} TFLAGS;

static struct {
	switch_hash_t *profile_hash;
	uint32_t callid;
} globals;

struct sipx_profile {
	int debug;
	char *name;
	char *dialplan;
	char *context;
	char *extrtpip;
	char *rtpip;
	char *sipip;
	char *extsipip;
	char *username;
	int tcpport;
	int udpport;
	int tlsport;
	char *codec_string;
	char *codec_order[SWITCH_MAX_CODECS];
	int codec_order_last;
	switch_hash_t *call_hash;
	int running;
	int codec_ms;
	int dtmf_duration;
	unsigned int flags;
	uint32_t max_calls;
};
struct private_object;

#define SipMessageSend(obj, msg) if (obj->profile->debug) {UtlString data;\
sipx_int_t len;\
msg.getBytes(&data,&len);\
 switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG,"Outbound Message\n%s\n", data.data());}\
obj->send(msg);

/* Our Decendant Classes ***************************************************/

class SipEndpoint: public SipUserAgent {
public:
	SipEndpoint(struct sipx_profile *nprofile): SipUserAgent(nprofile->tcpport,
															 nprofile->udpport,
															 nprofile->tlsport,
															 nprofile->extsipip,
															 nprofile->username,
															 nprofile->sipip) 
	{ 
		profile = nprofile; 
	}

	struct sipx_profile *profile;
	void SipEndpoint::handleInvite(const SipMessage *sipMsg);
	void SipEndpoint::handleInviteResponse(const SipMessage *sipMsg);
	void SipEndpoint::handleBye(const SipMessage *sipMsg);
	void SipEndpoint::sendBusy(const SipMessage *sipMsg);
	void SipEndpoint::sendBye(const SipMessage *sipMsg);
	void SipEndpoint::sendOk(const SipMessage *sipMsg);
	void SipEndpoint::sendNotFound(const SipMessage *sipMsg);
	void SipEndpoint::sendRinging(const SipMessage *sipMsg);
	void SipEndpoint::sendTrying(const SipMessage *sipMsg);
private:
	uint32_t calls;
};

class SipServer: public OsServerTask {
public:
	SipServer(struct sipx_profile *nprofile);
	~SipServer();
    virtual UtlBoolean handleMessage(OsMsg& eventMessage);
	SipEndpoint *endPoint;
private:
	OsMsgQ* queue;

};

SipServer::SipServer(struct sipx_profile *nprofile)
{
	endPoint = new SipEndpoint(nprofile);
	//SipEndpoint *endPoint = new SipEndpoint(profile);
	//queue = getMessageQueue();


	endPoint->addMessageObserver(*this->getMessageQueue(),
								 NULL,
								 TRUE,
								 TRUE,
								 TRUE,
								 FALSE);

	endPoint->allowMethod(SIP_INVITE_METHOD);
	endPoint->allowMethod(SIP_ACK_METHOD);
	endPoint->allowMethod(SIP_CANCEL_METHOD);
	endPoint->allowMethod(SIP_BYE_METHOD);
	endPoint->allowMethod(SIP_REFER_METHOD);
	endPoint->allowMethod(SIP_OPTIONS_METHOD);
	endPoint->start();
	start();


}

SipServer::~SipServer()
{
	delete endPoint;
}
/********************************************************************************/


struct private_object {
	uint32_t flags;
	switch_core_session_t *session;
	switch_frame_t read_frame;
	switch_codec_interface_t *codecs[SWITCH_MAX_CODECS];
	int num_codecs;
	switch_codec_t read_codec;
	switch_codec_t write_codec;
	uint32_t codec_index;
	uint32_t codec_rate;
	uint32_t codec_ms;
	switch_caller_profile_t *caller_profile;
	int32_t timestamp_send;
	int32_t timestamp_recv;
	int payload_num;
	switch_rtp_t *rtp_session;
	int ssrc;
	switch_time_t last_read;
	sipx_profile *profile;
	SipEndpoint *endPoint;
	const SipMessage *Anchor;
	char *local_sdp_audio_ip;
	switch_port_t local_sdp_audio_port;
	char *remote_sdp_audio_ip;
	switch_port_t remote_sdp_audio_port;
	char *adv_sdp_audio_ip;
	switch_port_t adv_sdp_audio_port;
	char *from_uri;
	char *to_uri;
	char *from_address;
	char *to_address;
	char *callid;
	char *far_end_contact;
	char *contact_url;
	SdpSrtpParameters srtpParams;
	switch_mutex_t *flag_mutex;
};



static switch_status_t sipx_on_init(switch_core_session_t *session);
static switch_status_t sipx_on_hangup(switch_core_session_t *session);
static switch_status_t sipx_on_loopback(switch_core_session_t *session);
static switch_status_t sipx_on_transmit(switch_core_session_t *session);
static switch_status_t sipx_outgoing_channel(switch_core_session_t *session, switch_caller_profile_t *outbound_profile,
											 switch_core_session_t **new_session, switch_memory_pool_t *pool);
static switch_status_t sipx_read_frame(switch_core_session_t *session, switch_frame_t **frame, int timeout,
									   switch_io_flag_t flags, int stream_id);
static switch_status_t sipx_write_frame(switch_core_session_t *session, switch_frame_t *frame, int timeout,
										switch_io_flag_t flags, int stream_id);
static int config_sipx(int reload);
static switch_status_t sipx_kill_channel(switch_core_session_t *session, int sig);
static switch_status_t activate_rtp(struct private_object *tech_pvt);
static void deactivate_rtp(struct private_object *tech_pvt);


static void tech_set_codecs(struct private_object *tech_pvt)
{
	if (tech_pvt->num_codecs) {
		return;
	}

	if (tech_pvt->profile->codec_string) {
		tech_pvt->num_codecs = switch_loadable_module_get_codecs_sorted(tech_pvt->codecs,
																		SWITCH_MAX_CODECS,
																		tech_pvt->profile->codec_order,
																		tech_pvt->profile->codec_order_last);
		
	} else {
		tech_pvt->num_codecs = switch_loadable_module_get_codecs(switch_core_session_get_pool(tech_pvt->session), tech_pvt->codecs,
																 sizeof(tech_pvt->codecs) / sizeof(tech_pvt->codecs[0]));
	}
}

static uint8_t tech_match_codec(struct private_object *tech_pvt, const SipMessage *sipMsg)
{
	uint8_t match = 0;
	UtlString rtp_ip;
	sipx_int_t numPayloadTypes = SWITCH_MAX_CODECS;
	sipx_int_t payloadTypes[SWITCH_MAX_CODECS];
	sipx_int_t pindex = -1, x;
	UtlString mediaType, test;
	sipx_int_t mediaPort;
	sipx_int_t mediaPortPairs;
	UtlString mediaTransportType;
	sipx_int_t len;
	const SdpBody *sdpBody = NULL;

	if (!(sdpBody = sipMsg->getSdpBody())) {
		return 0;
	}

	sdpBody->getBytes(&test, &len);
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "SDP\n%s\n", test.data());
	tech_set_codecs(tech_pvt);

	if ((pindex = sdpBody->findMediaType("audio", 0)) < 0) {
		return 0;
	}
	sdpBody->getMediaAddress(pindex, &rtp_ip);


	if (!sdpBody->getMediaData(pindex,
							  &mediaType,
							  &mediaPort,
							  &mediaPortPairs,
							  &mediaTransportType,
							  SWITCH_MAX_CODECS,
							  &numPayloadTypes,
							  payloadTypes)) {
		return 0;
	}

	

	for (x = 0; match == 0 && x < numPayloadTypes; x++) {
		UtlString mime;
		sipx_int_t rate, chans, i;

		sdpBody->getPayloadRtpMap (payloadTypes[x], mime, rate, chans);
		for (i = 0; i < tech_pvt->num_codecs; i++) {
			if (payloadTypes[x] > 96) {
				match = strcasecmp((char *)mime.data(), tech_pvt->codecs[i]->implementations->iananame) ? 0 : 1;
			} else {
				match = (payloadTypes[x] == tech_pvt->codecs[i]->implementations->ianacode) ? 1 : 0;
			}
						
			if (match) {
				const switch_codec_implementation_t *imp;
				match = 0;
				for (imp = tech_pvt->codecs[i]->implementations; imp; imp = imp->next) {
					if (imp->samples_per_second == (uint32_t)rate) {
						match = 1;
						break;
					}
				}
			}

			if (match) {
				tech_pvt->codec_index = i;
				tech_pvt->codec_rate = rate;
				tech_pvt->codec_ms = tech_pvt->profile->codec_ms;
				tech_pvt->remote_sdp_audio_ip = switch_core_session_strdup(tech_pvt->session, (char *)rtp_ip.data());
				tech_pvt->remote_sdp_audio_port = mediaPort;						
				break;
			}
		}
	}

	return match;
}


static switch_status_t tech_choose_port(struct private_object *tech_pvt)
{
	char *ip = tech_pvt->profile->rtpip;
	switch_port_t sdp_port;
	char *err;
	switch_channel_t *channel = switch_core_session_get_channel(tech_pvt->session);
	

	tech_pvt->local_sdp_audio_port = switch_rtp_request_port();
	sdp_port = tech_pvt->local_sdp_audio_port;


	if (tech_pvt->profile->extrtpip) {
		if (!strncasecmp(tech_pvt->profile->extrtpip, "stun:", 5)) {
			char *stun_ip = tech_pvt->profile->extrtpip + 5;
			if (!stun_ip) {
				switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Stun Failed! NO STUN SERVER\n");
				switch_channel_hangup(channel, SWITCH_CAUSE_DESTINATION_OUT_OF_ORDER);
				return SWITCH_STATUS_FALSE;
			}
			if (switch_stun_lookup(&ip,
								   &sdp_port,
								   stun_ip,
								   SWITCH_STUN_DEFAULT_PORT,
								   &err,
								   switch_core_session_get_pool(tech_pvt->session)) != SWITCH_STATUS_SUCCESS) {
				switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Stun Failed! %s:%d [%s]\n", stun_ip, SWITCH_STUN_DEFAULT_PORT, err);
				switch_channel_hangup(channel, SWITCH_CAUSE_DESTINATION_OUT_OF_ORDER);
				return SWITCH_STATUS_FALSE;
			}
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "Stun Success [%s]:[%d]\n", ip, sdp_port);
		} else {
			ip = tech_pvt->profile->extrtpip;
		}
	}

	tech_pvt->adv_sdp_audio_ip = switch_core_session_strdup(tech_pvt->session, ip);
	tech_pvt->adv_sdp_audio_port = sdp_port;
	return SWITCH_STATUS_SUCCESS;
}



/* 
State methods they get called when the state changes to the specific state 
returning SWITCH_STATUS_SUCCESS tells the core to execute the standard state method next
so if you fully implement the state you can return SWITCH_STATUS_FALSE to skip it.
*/
static switch_status_t sipx_on_init(switch_core_session_t *session)
{
	struct private_object *tech_pvt;
	switch_channel_t *channel = NULL;
	switch_caller_profile_t *caller_profile;
	SdpCodec *rtpCodecs[SWITCH_MAX_CODECS] = {0};
	SdpCodec staticRtpCodecs[SWITCH_MAX_CODECS];
	int x = 0, y = 0;

	channel = switch_core_session_get_channel(session);
	assert(channel != NULL);

	tech_pvt = (struct private_object *) switch_core_session_get_private(session);
	assert(tech_pvt != NULL);

	tech_pvt->read_frame.buflen = SWITCH_RTP_MAX_BUF_LEN;

	caller_profile = switch_channel_get_caller_profile(channel);

	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "SIPX INIT\n");

	if (switch_test_flag(tech_pvt, TFLAG_OUTBOUND)) {
		SipMessage Request;

		tech_set_codecs(tech_pvt);

		if (tech_pvt->num_codecs < 1 || tech_choose_port(tech_pvt) != SWITCH_STATUS_SUCCESS) {
			switch_channel_hangup(channel, SWITCH_CAUSE_DESTINATION_OUT_OF_ORDER);
            return SWITCH_STATUS_FALSE;
		}
		
		for (x = 0; (x < tech_pvt->num_codecs && y < SWITCH_MAX_CODECS); x++) {
			const switch_codec_implementation_t *imp;

			for (imp = tech_pvt->codecs[x]->implementations; imp; imp = imp->next) {
				staticRtpCodecs[y] = SdpCodec((SdpCodec::SdpCodecTypes)tech_pvt->codecs[x]->implementations->ianacode, 
											  0,
											  "audio",
											  tech_pvt->codecs[x]->implementations->iananame,
											  imp->samples_per_second,
											  imp->encoded_bytes_per_frame,
											  imp->number_of_channels
											  );
				rtpCodecs[y] = &staticRtpCodecs[y];
				y++;
			}
		}
		
		Request.setInviteData((const char *)tech_pvt->from_address,
							  (const char *)tech_pvt->to_address,
							  (const char *)tech_pvt->far_end_contact,
							  (const char *)tech_pvt->contact_url,
							  (const char *)tech_pvt->callid,
							  (const char *)tech_pvt->adv_sdp_audio_ip,
							  (int)tech_pvt->adv_sdp_audio_port,
							  0,
							  0,
							  0,
							  &tech_pvt->srtpParams,
							  1,
							  y,
							  rtpCodecs);

		SipMessageSend(tech_pvt->endPoint, Request);
		set_anchor(tech_pvt, Request);
	}


	/* Move Channel's State Machine to RING */
	switch_channel_set_state(channel, CS_RING);
	return SWITCH_STATUS_SUCCESS;
}

static switch_status_t sipx_on_ring(switch_core_session_t *session)
{
	switch_channel_t *channel = NULL;
	struct private_object *tech_pvt = NULL;

	channel = switch_core_session_get_channel(session);
	assert(channel != NULL);

	tech_pvt = (struct private_object *) switch_core_session_get_private(session);
	assert(tech_pvt != NULL);

	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "SIPX RING\n");

	return SWITCH_STATUS_SUCCESS;
}


static switch_status_t sipx_on_execute(switch_core_session_t *session)
{
	switch_channel_t *channel = NULL;
	struct private_object *tech_pvt = NULL;

	channel = switch_core_session_get_channel(session);
	assert(channel != NULL);

	tech_pvt = (struct private_object *) switch_core_session_get_private(session);
	assert(tech_pvt != NULL);

	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "SIPX EXECUTE\n");

	return SWITCH_STATUS_SUCCESS;
}

static switch_status_t sipx_on_hangup(switch_core_session_t *session)
{
	struct private_object *tech_pvt;
	switch_channel_t *channel = NULL;

	channel = switch_core_session_get_channel(session);
	assert(channel != NULL);

	tech_pvt = (struct private_object *) switch_core_session_get_private(session);
	assert(tech_pvt != NULL);

	deactivate_rtp(tech_pvt);

	if (switch_test_flag(tech_pvt, TFLAG_ANS) && !switch_test_flag(tech_pvt, TFLAG_BYE)) {
		tech_pvt->endPoint->sendBye(tech_pvt->Anchor);
		switch_set_flag_locked(tech_pvt, TFLAG_BYE);
	} else if (!switch_test_flag(tech_pvt, TFLAG_ANS)) {
		tech_pvt->endPoint->sendNotFound(tech_pvt->Anchor);
	}

	if (tech_pvt->Anchor) {
		delete tech_pvt->Anchor;
	}
	switch_core_hash_delete(tech_pvt->profile->call_hash, tech_pvt->callid);
	
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "SIPX HANGUP\n");

	return SWITCH_STATUS_SUCCESS;
}

static switch_status_t sipx_on_loopback(switch_core_session_t *session)
{
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "SIPX LOOPBACK\n");
	return SWITCH_STATUS_SUCCESS;
}

static switch_status_t sipx_on_transmit(switch_core_session_t *session)
{
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "SIPX TRANSMIT\n");
	return SWITCH_STATUS_SUCCESS;
}

static void deactivate_rtp(struct private_object *tech_pvt)
{
	int loops = 0;//, sock = -1;

	if (tech_pvt->rtp_session) {
		while (loops < 10 && (switch_test_flag(tech_pvt, TFLAG_READING) || switch_test_flag(tech_pvt, TFLAG_WRITING))) {
			switch_yield(10000);
			loops++;
		}
		switch_rtp_destroy(&tech_pvt->rtp_session);
		tech_pvt->rtp_session = NULL;
	}
}

static switch_status_t activate_rtp(struct private_object *tech_pvt)
{
	int bw, ms;
	switch_channel_t *channel;
	const char *err = NULL;
	switch_rtp_flag_t flags;

	assert(tech_pvt != NULL);

	channel = switch_core_session_get_channel(tech_pvt->session);
	assert(channel != NULL);

	assert(tech_pvt->codecs[tech_pvt->codec_index] != NULL);

	if (tech_pvt->rtp_session) {
	  return SWITCH_STATUS_SUCCESS;
	}


	if (switch_core_codec_init(&tech_pvt->read_codec,  
							   tech_pvt->codecs[tech_pvt->codec_index]->implementations->iananame,
							   tech_pvt->codec_rate,
							   tech_pvt->codec_ms,
							   1,
							   SWITCH_CODEC_FLAG_ENCODE | SWITCH_CODEC_FLAG_DECODE,
							   NULL,
							   switch_core_session_get_pool(tech_pvt->session)) != SWITCH_STATUS_SUCCESS) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Can't load codec?\n");
		switch_channel_hangup(channel, SWITCH_CAUSE_DESTINATION_OUT_OF_ORDER);
		return SWITCH_STATUS_FALSE;
	} else {
		if (switch_core_codec_init(&tech_pvt->write_codec,
								   tech_pvt->codecs[tech_pvt->codec_index]->implementations->iananame,
								   tech_pvt->codec_rate,
								   tech_pvt->codec_ms,
								   1,
								   SWITCH_CODEC_FLAG_ENCODE | SWITCH_CODEC_FLAG_DECODE,
								   NULL,
								   switch_core_session_get_pool(tech_pvt->session)) != SWITCH_STATUS_SUCCESS) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Can't load codec?\n");
			switch_channel_hangup(channel, SWITCH_CAUSE_DESTINATION_OUT_OF_ORDER);
			return SWITCH_STATUS_FALSE;
		} else {
			int ms;
			tech_pvt->read_frame.rate = tech_pvt->codec_rate;
			ms = tech_pvt->write_codec.implementation->microseconds_per_frame / 1000;
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "Activate Inbound Codec %s/%d %d ms\n",
							  tech_pvt->codecs[tech_pvt->codec_index]->implementations->iananame, tech_pvt->codec_rate, tech_pvt->codec_ms);
			tech_pvt->read_frame.codec = &tech_pvt->read_codec;
			switch_core_session_set_read_codec(tech_pvt->session, &tech_pvt->read_codec);
			switch_core_session_set_write_codec(tech_pvt->session, &tech_pvt->write_codec);
		}
	}

	
	bw = tech_pvt->read_codec.implementation->bits_per_second;
	ms = tech_pvt->read_codec.implementation->microseconds_per_frame;

	flags = (switch_rtp_flag_t) (SWITCH_RTP_FLAG_RAW_WRITE | SWITCH_RTP_FLAG_MINI);
	if (switch_test_flag(tech_pvt, TFLAG_TIMER)) {
	  flags = (switch_rtp_flag_t) (flags | SWITCH_RTP_FLAG_USE_TIMER | SWITCH_RTP_FLAG_TIMER_RECLOCK);
	}
	
	tech_pvt->rtp_session = switch_rtp_new(tech_pvt->local_sdp_audio_ip,
					       tech_pvt->local_sdp_audio_port,
					       tech_pvt->remote_sdp_audio_ip,
					       tech_pvt->remote_sdp_audio_port,
					       tech_pvt->read_codec.implementation->ianacode,
					       tech_pvt->read_codec.implementation->encoded_bytes_per_frame,
					       tech_pvt->codec_ms * 1000,
					       (switch_rtp_flag_t)
					       flags,
					       NULL,
					       &err,
					       switch_core_session_get_pool(tech_pvt->session));
	
	if (tech_pvt->rtp_session) {
		uint8_t vad_in = switch_test_flag(tech_pvt, TFLAG_VAD_IN) ? 1 : 0;
		uint8_t vad_out = switch_test_flag(tech_pvt, TFLAG_VAD_OUT) ? 1 : 0;
		uint8_t inb = switch_test_flag(tech_pvt, TFLAG_OUTBOUND) ? 0 : 1;

		tech_pvt->ssrc = switch_rtp_get_ssrc(tech_pvt->rtp_session);
		switch_set_flag_locked(tech_pvt, TFLAG_RTP);
		
		if ((vad_in && inb) || (vad_out && !inb)) {
			switch_rtp_enable_vad(tech_pvt->rtp_session, tech_pvt->session, &tech_pvt->read_codec, SWITCH_VAD_FLAG_TALKING);
			switch_set_flag_locked(tech_pvt, TFLAG_VAD);
		}
	} else {
		switch_channel_t *channel = switch_core_session_get_channel(tech_pvt->session);
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "RTP REPORTS ERROR: [%s]\n", err);
		switch_channel_hangup(channel, SWITCH_CAUSE_DESTINATION_OUT_OF_ORDER);
		switch_clear_flag_locked(tech_pvt, TFLAG_IO);
		return SWITCH_STATUS_FALSE;
	}
		
	switch_set_flag_locked(tech_pvt, TFLAG_IO);
	return SWITCH_STATUS_SUCCESS;
}

static switch_status_t sipx_answer_channel(switch_core_session_t *session)
{
	struct private_object *tech_pvt;
	switch_channel_t *channel = NULL;

	
	assert(session != NULL);

	channel = switch_core_session_get_channel(session);
	assert(channel != NULL);

	tech_pvt = (struct private_object *) switch_core_session_get_private(session);
	assert(tech_pvt != NULL);

	if (!switch_test_flag(tech_pvt, TFLAG_ANS) && !switch_channel_test_flag(channel, CF_OUTBOUND)) {
		SipMessage Response;
		SdpCodec *rtpCodecs[1] = {0};
		SdpCodec Codec;

		switch_set_flag_locked(tech_pvt, TFLAG_ANS);

		Codec = SdpCodec((SdpCodec::SdpCodecTypes)tech_pvt->codecs[tech_pvt->codec_index]->implementations->ianacode, 
						 0,
						 "audio",
						 tech_pvt->codecs[tech_pvt->codec_index]->implementations->iananame,
						 tech_pvt->read_codec.implementation->samples_per_second,
						 tech_pvt->read_codec.implementation->encoded_bytes_per_frame,
						 tech_pvt->read_codec.implementation->number_of_channels
						 );

		rtpCodecs[0] = &Codec;


		Response.setInviteOkData(
								 tech_pvt->Anchor,
								 (const char *)tech_pvt->adv_sdp_audio_ip,
								 (int)tech_pvt->adv_sdp_audio_port,
								 0,
								 0,
								 0,
								 1,
								 rtpCodecs,
								 tech_pvt->srtpParams,
								 300,
								 tech_pvt->to_address
								 );

		SipMessageSend(tech_pvt->endPoint, Response);
	}

	return SWITCH_STATUS_SUCCESS;
}


static switch_status_t sipx_read_frame(switch_core_session_t *session, switch_frame_t **frame, int timeout,
									   switch_io_flag_t flags, int stream_id)
{
	struct private_object *tech_pvt = NULL;
	size_t bytes = 0, samples = 0, frames = 0, ms = 0;
	switch_channel_t *channel = NULL;
	int payload = 0;
	switch_time_t now, started = switch_time_now(), last_act = switch_time_now();
	unsigned int elapsed;
	uint32_t hard_timeout = 60000 * 3;

	channel = switch_core_session_get_channel(session);
	assert(channel != NULL);

	tech_pvt = (struct private_object *) switch_core_session_get_private(session);
	assert(tech_pvt != NULL);

	if (switch_test_flag(tech_pvt, TFLAG_HUP)) {
		return SWITCH_STATUS_FALSE;
	}

	tech_pvt->read_frame.datalen = 0;
	switch_set_flag_locked(tech_pvt, TFLAG_READING);


	bytes = tech_pvt->read_codec.implementation->encoded_bytes_per_frame;
	samples = tech_pvt->read_codec.implementation->samples_per_frame;
	ms = tech_pvt->read_codec.implementation->microseconds_per_frame;
	
	if (tech_pvt->last_read) {
		elapsed = (unsigned int)((switch_time_now() - tech_pvt->last_read) / 1000);
		if (elapsed > 60000) {
			return SWITCH_STATUS_TIMEOUT;
		}
	}

	if (switch_test_flag(tech_pvt, TFLAG_IO)) {
		switch_status_t status;

		if (!switch_test_flag(tech_pvt, TFLAG_RTP)) {
			return SWITCH_STATUS_GENERR;
		}

		assert(tech_pvt->rtp_session != NULL);
		tech_pvt->read_frame.datalen = 0;


		while (!switch_test_flag(tech_pvt, TFLAG_BYE) && switch_test_flag(tech_pvt, TFLAG_IO) && tech_pvt->read_frame.datalen == 0) {
			now = switch_time_now();
			tech_pvt->read_frame.flags = SFF_NONE;

			status = switch_rtp_zerocopy_read_frame(tech_pvt->rtp_session, &tech_pvt->read_frame);
			if (status != SWITCH_STATUS_SUCCESS && status != SWITCH_STATUS_BREAK) {
				return SWITCH_STATUS_FALSE;
			}

			
			
			payload = tech_pvt->read_frame.payload;


			elapsed = (unsigned int)((switch_time_now() - started) / 1000);

			if (timeout > -1) {
				if (elapsed >= (unsigned int)timeout) {
					return SWITCH_STATUS_BREAK;
				}
			}
			
			elapsed = (unsigned int)((switch_time_now() - last_act) / 1000);
			if (elapsed >= hard_timeout) {
				return SWITCH_STATUS_BREAK;
			}

			if (switch_rtp_has_dtmf(tech_pvt->rtp_session)) {
				char dtmf[128];
				switch_rtp_dequeue_dtmf(tech_pvt->rtp_session, dtmf, sizeof(dtmf));
				switch_channel_queue_dtmf(channel, dtmf);
			}


			if (tech_pvt->read_frame.datalen > 0) {
				tech_pvt->last_read = switch_time_now();
				bytes = tech_pvt->read_codec.implementation->encoded_bytes_per_frame;
				frames = (tech_pvt->read_frame.datalen / bytes);
				samples = frames * tech_pvt->read_codec.implementation->samples_per_frame;
				ms = frames * tech_pvt->read_codec.implementation->microseconds_per_frame;
				tech_pvt->timestamp_recv += (int32_t) samples;
				tech_pvt->read_frame.samples = (int) samples;
				break;
			}

			switch_yield(1000);
		}

	} 

	switch_clear_flag_locked(tech_pvt, TFLAG_READING);

	*frame = &tech_pvt->read_frame;

	return SWITCH_STATUS_SUCCESS;
}


static switch_status_t sipx_write_frame(switch_core_session_t *session, switch_frame_t *frame, int timeout,
										switch_io_flag_t flags, int stream_id)
{
	struct private_object *tech_pvt;
	switch_channel_t *channel = NULL;
	switch_status_t status = SWITCH_STATUS_SUCCESS;
	int bytes = 0, samples = 0, frames = 0;

	channel = switch_core_session_get_channel(session);
	assert(channel != NULL);

	tech_pvt = (struct private_object *) switch_core_session_get_private(session);
	assert(tech_pvt != NULL);

	if (switch_test_flag(tech_pvt, TFLAG_HUP)) {
		return SWITCH_STATUS_FALSE;
	}

	if (!switch_test_flag(tech_pvt, TFLAG_RTP)) {
		return SWITCH_STATUS_GENERR;
	}

	if (!switch_test_flag(tech_pvt, TFLAG_IO)) {
		return SWITCH_STATUS_SUCCESS;
	}

	switch_set_flag_locked(tech_pvt, TFLAG_WRITING);


	bytes = tech_pvt->read_codec.implementation->encoded_bytes_per_frame;
	frames = ((int) frame->datalen / bytes);
	samples = frames * tech_pvt->read_codec.implementation->samples_per_frame;


	//printf("%s %s->%s send %d bytes %d samples in %d frames ts=%d\n", switch_channel_get_name(channel), tech_pvt->local_sdp_audio_ip, tech_pvt->remote_sdp_audio_ip, frame->datalen, samples, frames, tech_pvt->timestamp_send);

	switch_rtp_write_frame(tech_pvt->rtp_session, frame, samples);
	
	tech_pvt->timestamp_send += (int) samples;

	switch_clear_flag_locked(tech_pvt, TFLAG_WRITING);
	return status;
}



static switch_status_t sipx_kill_channel(switch_core_session_t *session, int sig)
{
	struct private_object *tech_pvt;
	switch_channel_t *channel = NULL;

	channel = switch_core_session_get_channel(session);
	assert(channel != NULL);

	tech_pvt = (struct private_object *) switch_core_session_get_private(session);
	assert(tech_pvt != NULL);

	switch_clear_flag_locked(tech_pvt, TFLAG_IO);
	switch_set_flag_locked(tech_pvt, TFLAG_HUP);

	if (tech_pvt->rtp_session) {
		switch_rtp_kill_socket(tech_pvt->rtp_session);
	}

	return SWITCH_STATUS_SUCCESS;

}

static switch_status_t sipx_waitfor_read(switch_core_session_t *session, int ms, int stream_id)
{
	struct private_object *tech_pvt;
	switch_channel_t *channel = NULL;

	channel = switch_core_session_get_channel(session);
	assert(channel != NULL);

	tech_pvt = (struct private_object *) switch_core_session_get_private(session);
	assert(tech_pvt != NULL);

	return SWITCH_STATUS_SUCCESS;
}


static switch_status_t sipx_waitfor_write(switch_core_session_t *session, int ms, int stream_id)
{
	struct private_object *tech_pvt;
	switch_channel_t *channel = NULL;

	channel = switch_core_session_get_channel(session);
	assert(channel != NULL);

	tech_pvt = (struct private_object *) switch_core_session_get_private(session);
	assert(tech_pvt != NULL);

	return SWITCH_STATUS_SUCCESS;

}

static switch_status_t sipx_send_dtmf(switch_core_session_t *session, char *digits)
{
	struct private_object *tech_pvt;

	tech_pvt = (struct private_object *) switch_core_session_get_private(session);
    assert(tech_pvt != NULL);

	return switch_rtp_queue_rfc2833(tech_pvt->rtp_session,
									digits,
									tech_pvt->profile->dtmf_duration * (tech_pvt->read_codec.implementation->samples_per_second / 1000));
	
}

static switch_status_t sipx_receive_message(switch_core_session_t *session, switch_core_session_message_t *msg)
{
	switch_channel_t *channel;
	struct private_object *tech_pvt;
			
	channel = switch_core_session_get_channel(session);
	assert(channel != NULL);
			
	tech_pvt = (struct private_object *) switch_core_session_get_private(session);
	assert(tech_pvt != NULL);


	switch (msg->message_id) {
	case SWITCH_MESSAGE_INDICATE_BRIDGE:
	  if (tech_pvt->rtp_session && switch_test_flag(tech_pvt, TFLAG_TIMER)) {
	    switch_rtp_clear_flag(tech_pvt->rtp_session, SWITCH_RTP_FLAG_USE_TIMER);
	    switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "De-activate timed RTP!\n");
	  }
	  break;
	case SWITCH_MESSAGE_INDICATE_UNBRIDGE:
	  if (tech_pvt->rtp_session && switch_test_flag(tech_pvt, TFLAG_TIMER)) {
	    switch_rtp_set_flag(tech_pvt->rtp_session, SWITCH_RTP_FLAG_USE_TIMER);
	    switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Re-activate timed RTP!\n");
	  }
	  break;
	case SWITCH_MESSAGE_INDICATE_PROGRESS: {
		SipMessage Response;
		SdpCodec *rtpCodecs[1] = {0};
		SdpCodec Codec;

		Codec = SdpCodec((SdpCodec::SdpCodecTypes)tech_pvt->codecs[tech_pvt->codec_index]->implementations->ianacode, 
						 0,
						 "audio",
						 tech_pvt->codecs[tech_pvt->codec_index]->implementations->iananame,
						 tech_pvt->read_codec.implementation->samples_per_second,
						 tech_pvt->read_codec.implementation->encoded_bytes_per_frame,
						 tech_pvt->read_codec.implementation->number_of_channels
						 );

		rtpCodecs[0] = &Codec;


		Response.setResponseData(tech_pvt->Anchor,
								 SIP_EARLY_MEDIA_CODE,
								 SIP_EARLY_MEDIA_TEXT);

		Response.addSdpBody(
							(const char *)tech_pvt->adv_sdp_audio_ip,
							(int)tech_pvt->adv_sdp_audio_port,
							0,
							0,
							0,
							1,
							rtpCodecs,
							&tech_pvt->srtpParams
							);

		SipMessageSend(tech_pvt->endPoint, Response);
	}
		break;
	default:
		break;
	}

	return SWITCH_STATUS_SUCCESS;
}

static const switch_io_routines_t sipx_io_routines = {
	/*.outgoing_channel */ sipx_outgoing_channel,
	/*.answer_channel */ sipx_answer_channel,
	/*.read_frame */ sipx_read_frame,
	/*.write_frame */ sipx_write_frame,
	/*.kill_channel */ sipx_kill_channel,
	/*.waitfor_read */ sipx_waitfor_read,
	/*.waitfor_read */ sipx_waitfor_write,
	/*.send_dtmf*/ sipx_send_dtmf,
	/*.receive_message*/ sipx_receive_message
};

static const switch_state_handler_table_t sipx_event_handlers = {
	/*.on_init */ sipx_on_init,
	/*.on_ring */ sipx_on_ring,
	/*.on_execute */ sipx_on_execute,
	/*.on_hangup */ sipx_on_hangup,
	/*.on_loopback */ sipx_on_loopback,
	/*.on_transmit */ sipx_on_transmit
};

static const switch_endpoint_interface_t sipx_endpoint_interface = {
	/*.interface_name */ "sipx",
	/*.io_routines */ &sipx_io_routines,
	/*.event_handlers */ &sipx_event_handlers,
	/*.private */ NULL,
	/*.next */ NULL
};

static const switch_loadable_module_interface_t sipx_module_interface = {
	/*.module_name */ modname,
	/*.endpoint_interface */ &sipx_endpoint_interface,
	/*.timer_interface */ NULL,
	/*.dialplan_interface */ NULL,
	/*.codec_interface */ NULL,
	/*.application_interface */ NULL
};

static switch_status_t sipx_outgoing_channel(switch_core_session_t *session, switch_caller_profile_t *outbound_profile,
											 switch_core_session_t **new_session, switch_memory_pool_t *pool)
{
	struct private_object *tech_pvt;
	switch_channel_t *channel;
	char name[128];
	switch_caller_profile_t *caller_profile = NULL;
	char workspace[1024];
	char *profile_name = NULL;
	char *proto_name = NULL;
	char *dest = NULL;
	SipServer *Server;

	if (!outbound_profile) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Doh! no caller profile\n");
		switch_core_session_destroy(new_session);
		return SWITCH_STATUS_GENERR;
	}


	switch_copy_string(workspace, outbound_profile->destination_number, sizeof(workspace));
	profile_name = workspace;
	if ((proto_name = strchr(profile_name, '/'))) {
		*proto_name++ = '\0';
		if ((dest = strchr(proto_name, '/'))) {
			*dest++ = '\0';
		}
	}

	if (!(profile_name && proto_name && dest)) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Bad URL!\n");
		return SWITCH_STATUS_GENERR;
	}

	outbound_profile->destination_number += strlen(profile_name) + strlen(proto_name) + 2;

	if (!(*new_session = switch_core_session_request(&sipx_endpoint_interface, pool)) != 0) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Session Request Failed!\n");
		return SWITCH_STATUS_GENERR;
	}

	if (!(tech_pvt = (struct private_object *) switch_core_session_alloc(*new_session, sizeof(struct private_object))) != 0) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_CRIT, "Hey where is my memory pool?\n");
		switch_core_session_destroy(new_session);
		return SWITCH_STATUS_GENERR;
	}
	memset(tech_pvt, 0, sizeof(*tech_pvt));
	switch_mutex_init(&tech_pvt->flag_mutex, SWITCH_MUTEX_NESTED, switch_core_session_get_pool(*new_session));
	
	if (!(Server = (SipServer *) switch_core_hash_find(globals.profile_hash, profile_name))) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Unknown Profile [%s]!\n", profile_name);
		return SWITCH_STATUS_GENERR;
	}
	tech_pvt->endPoint = Server->endPoint;

	tech_pvt->profile = tech_pvt->endPoint->profile;
	tech_pvt->flags = tech_pvt->profile->flags;
	assert(tech_pvt->profile != NULL);
	tech_pvt->to_address = switch_core_session_strdup(*new_session, dest);
	dest = NULL;
	profile_name = NULL;
	proto_name = NULL;
	
	snprintf(workspace, sizeof(workspace), "sip:%s@%s", outbound_profile->caller_id_number, tech_pvt->profile->sipip);
	tech_pvt->contact_url = switch_core_session_strdup(*new_session, workspace);
	snprintf(workspace, sizeof(workspace), "sip:%s@%s", tech_pvt->profile->username, tech_pvt->profile->sipip);
	tech_pvt->from_address = switch_core_session_strdup(*new_session, workspace);
	//tech_pvt->far_end_contact = tech_pvt->from_address;
	//Remote-Party-ID: "Brian West" <sip:1000@bkw.org>;privacy=off;

	tech_pvt->callid = switch_core_session_strdup(*new_session, switch_core_session_get_uuid(*new_session));
	switch_core_hash_insert(tech_pvt->profile->call_hash, tech_pvt->callid, tech_pvt);
	switch_core_session_add_stream(*new_session, NULL);
	
	
	channel = switch_core_session_get_channel(*new_session);
	switch_core_session_set_private(*new_session, tech_pvt);
	tech_pvt->session = *new_session;


	snprintf(name, sizeof(name), "sipX/%s-%04x", outbound_profile->destination_number, rand() & 0xffff);
	switch_channel_set_name(channel, name);
	caller_profile = switch_caller_profile_clone(*new_session, outbound_profile);
	switch_channel_set_caller_profile(channel, caller_profile);
	tech_pvt->caller_profile = caller_profile;
	switch_channel_set_flag(channel, CF_OUTBOUND);
	switch_set_flag_locked(tech_pvt, TFLAG_OUTBOUND);
	switch_channel_set_state(channel, CS_INIT);


	return SWITCH_STATUS_SUCCESS;
}


SWITCH_MOD_DECLARE(switch_status_t) switch_module_shutdown(void)
{
	return SWITCH_STATUS_SUCCESS;
}

static int config_sipx(int reload)
{
	char *cf = "sipx.conf";
	switch_xml_t cfg, xml, xprofile, param;
	
	if (!(xml = switch_xml_open_cfg(cf, &cfg, NULL))) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "open of %s failed\n", cf);
		return SWITCH_STATUS_TERM;
	}

	for (xprofile = switch_xml_child(cfg, "profile"); xprofile; xprofile = xprofile->next) {
		char *xprofilename = (char *) switch_xml_attr_soft(xprofile, "name");
		struct sipx_profile *profile;

		if (!(profile = (struct sipx_profile *) malloc(sizeof(*profile)))) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Memory Error!\n");
			continue;
		}
		memset(profile, 0, sizeof(*profile));

		if (!xprofilename) {
			xprofilename = "unnamed";
		}

		set_param(profile->name, xprofilename);

		profile->dtmf_duration = 100;		
		for (param = switch_xml_child(xprofile, "param"); param; param = param->next) {
			char *var = (char *) switch_xml_attr_soft(param, "name");
			char *val = (char *) switch_xml_attr_soft(param, "value");

			if (!strcmp(var, "debug")) {
				profile->debug = atoi(val);
			} else if (!strcmp(var, "use-rtp-timer") && switch_true(val)) {
			  	switch_set_flag(profile, TFLAG_TIMER);
			} else if (!strcmp(var, "tcp-port")) {
				profile->tcpport = atoi(val);
			} else if (!strcmp(var, "udp-port")) {
				profile->udpport = atoi(val);
			} else if (!strcmp(var, "tls-port")) {
				profile->tlsport = atoi(val);
			} else if (!strcmp(var, "vad")) {
				if (!strcasecmp(val, "in")) {
					switch_set_flag(profile, TFLAG_VAD_IN);
				} else if (!strcasecmp(val, "out")) {
					switch_set_flag(profile, TFLAG_VAD_OUT);
				} else if (!strcasecmp(val, "both")) {
					switch_set_flag(profile, TFLAG_VAD_IN);
					switch_set_flag(profile, TFLAG_VAD_OUT);
				} else {
					switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Invald option %s for VAD\n", val);
				}
			} else if (!strcmp(var, "ext-rtp-ip")) {
				set_param(profile->extrtpip, val);
			} else if (!strcmp(var, "rtp-ip")) {
				set_param(profile->rtpip, val);
			} else if (!strcmp(var, "sip-ip")) {
				set_param(profile->sipip, val);
			} else if (!strcmp(var, "ext-sip-ip")) {
				set_param(profile->extsipip, val);
			} else if (!strcmp(var, "username")) {
				set_param(profile->username, val);
			} else if (!strcmp(var, "context")) {
				set_param(profile->context, val);
			} else if (!strcmp(var, "dialplan")) {
				set_param(profile->dialplan, val);
			} else if (!strcmp(var, "max-calls")) {
				profile->max_calls = atoi(val);
			} else if (!strcmp(var, "codec-prefs")) {
				set_param(profile->codec_string, val);
				profile->codec_order_last = switch_separate_string(profile->codec_string, ',', profile->codec_order, SWITCH_MAX_CODECS);
			} else if (!strcmp(var, "codec-ms")) {
				profile->codec_ms = atoi(val);
			} else if (!strcmp(var, "dtmf-duration")) {
				int dur = atoi(val);
				if (dur > 10 && dur < 8000) {
					profile->dtmf_duration = dur;
				} else {
					switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Duration out of bounds!\n");
				}
			}
		}

		if (!profile->username) {
			set_param(profile->username, "FreeSWITCH");
		}

		if (!profile->rtpip) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_WARNING, "Setting ip to '127.0.0.1'\n");
			set_param(profile->rtpip, "127.0.0.1");
		}

		if (!profile->codec_ms) {
			profile->codec_ms = 20;
		}
		
		if (!profile->udpport) {
			profile->udpport = 5060;
		}

		if (!profile->dialplan) {
			set_param(profile->dialplan, "default");
		}

		if (profile) {
			SipServer *Server = new SipServer(profile);
			switch_core_hash_init(&profile->call_hash, module_pool);
			switch_core_hash_insert(globals.profile_hash, profile->name, Server);
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "Started Profile %s [%s]\n", profile->name, profile->sipip ? profile->sipip : "any");
		}
	}
	
	switch_xml_free(xml);

	return 0;

}


SWITCH_MOD_DECLARE(switch_status_t) switch_module_load(const switch_loadable_module_interface_t **module_interface, char *filename)
{

	/* NOTE:  **interface is **_interface because the common lib redefines interface to struct in some situations */

	if (switch_core_new_memory_pool(&module_pool) != SWITCH_STATUS_SUCCESS) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_CRIT, "OH OH no pool\n");
		return SWITCH_STATUS_TERM;
	}

	switch_core_hash_init(&globals.profile_hash, module_pool);

	config_sipx(0);

	/* connect my internal structure to the blank pointer passed to me */
	*module_interface = &sipx_module_interface;

	/* indicate that the module should continue to be loaded */
	return SWITCH_STATUS_SUCCESS;

}


/* Our Decendant Class ***************************************************/



void SipEndpoint::sendBye(const SipMessage *sipMsg)
{
	SipMessage Response;
	Response.setByeData(sipMsg, NULL, FALSE, 1, NULL, NULL, NULL);
	SipMessageSend(this, Response);
}

void SipEndpoint::sendOk(const SipMessage *sipMsg)
{
	SipMessage Response;
	Response.setResponseData(sipMsg,
							 SIP_OK_CODE,
							 SIP_OK_TEXT);
	SipMessageSend(this, Response);
}

void SipEndpoint::sendBusy(const SipMessage *sipMsg)
{
	SipMessage Response;
	Response.setInviteBusyData(sipMsg);
	SipMessageSend(this, Response);
}

void SipEndpoint::sendNotFound(const SipMessage *sipMsg)
{

	return sendBusy(sipMsg);
}

void SipEndpoint::sendRinging(const SipMessage *sipMsg)
{
	SipMessage Response;
	Response.setInviteRingingData(sipMsg);
	SipMessageSend(this, Response);
}

void SipEndpoint::sendTrying(const SipMessage *sipMsg)
{
	SipMessage Response;
	Response.setTryingResponseData(sipMsg);
	SipMessageSend(this, Response);
}

void SipEndpoint::handleBye(const SipMessage *sipMsg)
{
	UtlString callid;
	struct private_object *tech_pvt;

	sipMsg->getCallIdField(&callid);
	if ((tech_pvt = (struct private_object *) switch_core_hash_find(profile->call_hash, (char *)callid.data()))) {
		switch_channel_t *channel = switch_core_session_get_channel(tech_pvt->session);
		switch_clear_flag_locked(tech_pvt, TFLAG_IO);
		switch_set_flag_locked(tech_pvt, TFLAG_BYE);
		switch_channel_hangup(channel, SWITCH_CAUSE_NORMAL_CLEARING);
	}

	sendOk(sipMsg);
}

void SipEndpoint::handleInviteResponse(const SipMessage *sipMsg)
{
	uint32_t responseCode = sipMsg->getResponseStatusCode();
	struct private_object *tech_pvt;
	UtlString callid;
	switch_channel_t *channel;


	sipMsg->getCallIdField(&callid);
	if (!(tech_pvt = (struct private_object *) switch_core_hash_find(profile->call_hash, (char *)callid.data()))) {
		sendBye(sipMsg);
		return;
	}
	
	channel = switch_core_session_get_channel(tech_pvt->session);

	switch (responseCode) {
	case SIP_OK_CODE: {
		SipMessage Response;
		int match;

		if ((match = tech_match_codec(tech_pvt, sipMsg))) {
			activate_rtp(tech_pvt);
			switch_channel_answer(channel);
			Response.setAckData(sipMsg, tech_pvt->Anchor);

			set_anchor(tech_pvt, *sipMsg);
			SipMessageSend(this, Response);
		} else {
			switch_channel_hangup(channel, SWITCH_CAUSE_DESTINATION_OUT_OF_ORDER);
			return;
		}
	}
		break;
	}
}


void SipEndpoint::handleInvite(const SipMessage *sipMsg)
{
	switch_core_session_t *session;
	switch_channel_t *channel;
	struct private_object *tech_pvt;
	char name[256] = "";
	UtlString displayname, protocol, username, ip, callid, remoteAddress;
	UtlString to_ip, to_protocol, to_username;
	sipx_int_t port, to_port, match = 0;


	if (profile->max_calls && (calls + 1) > profile->max_calls) {
		return sendBusy(sipMsg);
	}

	if (!(session = switch_core_session_request(&sipx_endpoint_interface, NULL))) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Memory Error!\n");
		return sendNotFound(sipMsg);
	}
	

	switch_core_session_add_stream(session, NULL);
	
	if (!(tech_pvt = (struct private_object *) switch_core_session_alloc(session, sizeof(struct private_object)))) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_CRIT, "Hey where is my memory pool?\n");
		switch_core_session_destroy(&session);
		return sendNotFound(sipMsg);
	}

	memset(tech_pvt, 0, sizeof(*tech_pvt));
	switch_mutex_init(&tech_pvt->flag_mutex, SWITCH_MUTEX_NESTED, switch_core_session_get_pool(session));
	tech_pvt->flags = profile->flags;
	channel = switch_core_session_get_channel(session);
	switch_core_session_set_private(session, tech_pvt);
	tech_pvt->session = session;
	tech_pvt->profile = profile;
	tech_pvt->endPoint = this;
	switch_set_flag_locked(tech_pvt, TFLAG_INBOUND);

	sipMsg->getFromUri(&username);
	snprintf(name, sizeof(name), "sipX/%s-%04x", username.data(), rand() & 0xffff);
	switch_channel_set_name(channel, name);

	sipMsg->getFromLabel(&displayname);
	displayname.strip(UtlString::both, '"');

	sipMsg->getFromAddress(&ip, &port, &protocol, &username);
	sipMsg->getCallIdField(&callid);
	
	tech_pvt->callid = switch_core_strdup(module_pool, (char *)callid.data());
	switch_core_hash_insert(profile->call_hash, tech_pvt->callid, tech_pvt);
	sipMsg->getToAddress(&to_ip, &to_port, &to_protocol, &to_username);

	sipMsg->getFromField(&callid);
	tech_pvt->from_address = switch_core_strdup(module_pool, (char *)callid.data());
	sipMsg->getToField(&callid);
	tech_pvt->to_address = switch_core_strdup(module_pool, (char *)callid.data());
	sipMsg->getFromUri(&callid);
	tech_pvt->from_uri = switch_core_strdup(module_pool, (char *)callid.data());
	sipMsg->getToUri(&callid);
	tech_pvt->to_uri = switch_core_strdup(module_pool, (char *)callid.data());

	set_anchor(tech_pvt, *sipMsg);

	if ((tech_pvt->caller_profile = switch_caller_profile_new(switch_core_session_get_pool(session),
								  (char *)to_username.data(),
								  profile->dialplan,
								  (char *)displayname.data(),
								  (char *)username.data(),
								  (char *)ip.data(),
								  NULL,
								  NULL,
								  NULL,
								  (char *)modname,
								  profile->context,
								  (char *)to_username.data())) != 0) {
		switch_channel_set_caller_profile(channel, tech_pvt->caller_profile);
	}

	if ((match = tech_match_codec(tech_pvt, sipMsg))) {
		if (tech_choose_port(tech_pvt) != SWITCH_STATUS_SUCCESS) {
			return sendNotFound(sipMsg);
		}
		activate_rtp(tech_pvt);
		switch_channel_set_state(channel, CS_INIT);
		switch_core_session_thread_launch(session);
	} else {
		sendNotFound(sipMsg);
	}
}

UtlBoolean SipServer::handleMessage(OsMsg& eventMessage)
{
    sipx_int_t msgType = eventMessage.getMsgType();
    sipx_int_t msgSubType = eventMessage.getMsgSubType();
	sipx_int_t seq;
	UtlBoolean ret = TRUE;

    if ((msgType == OsMsg::PHONE_APP) && (msgSubType == SipMessage::NET_SIP_MESSAGE)) {
		const SipMessage* sipMsg = ((SipMessageEvent&)eventMessage).getMessage();
		UtlString method;

		if (endPoint->profile->debug) {
				UtlString data;
				sipx_int_t len;
				sipMsg->getBytes(&data,&len);
				switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG,"Inbound Message\n%s\n", data.data());
		}
		
		
		if (sipMsg->getCSeqField(&seq, &method)) {
			if (sipMsg->isResponse()) {
				if (0==method.compareTo(SIP_INVITE_METHOD, UtlString::ignoreCase)) {
					endPoint->handleInviteResponse(sipMsg);
				} else {
					switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Unhandled Response to [%s]\n", method.data());
				}
			} else {
				if (0==method.compareTo(SIP_INVITE_METHOD, UtlString::ignoreCase)) {
					endPoint->handleInvite(sipMsg);
				} else if (0==method.compareTo(SIP_BYE_METHOD, UtlString::ignoreCase)) {
					endPoint->handleBye(sipMsg);
				} else {
					switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Unhandled Event [%s]\n", method.data());
				}
			}
		}
	}
	

	return ret;
}


